/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// d_part.c: software driver module for drawing particles

#include "quakedef.h"
#include "d_local.h"

// not used by software driver
int	d_pix_shift;

// Manoel Kasimier - begin
typedef struct d_particle_s
{
	int
		h_radius
	,	v_radius
	,	vrectright
	,	vrectbottom
	,	zcenter
	,	lines
	,	*scanwidth
	,	*offset
		;
	particle_t
		*unit
		;
} d_particle_t;

typedef struct r_particles_s
{
	int
		max_size
		;
	d_particle_t
		**for_size
		;
	float
		scale
		;
	double
		h_aspect
		;
} r_particles_t;

r_particles_t	r_particles;

qboolean		r_particles_init = false;

void D_Particles_Recalc (void)
{
	int
		p = 0
	,	max_size
		;
	double
		h_aspect
	,	pix
	,	width
	,	height
	,	count
	,	offset
		;

	r_particles.scale = fovscale * ( (float) r_refdef.vrect.height / 60.0f); // 30.0f = 240.0f / 8.0f

	h_aspect = 1.0 / (double) vid.aspect;

	// doubled the minimum size here to make sure the particle will appear,
	// since the width and height calculations reduces its drawing area
	max_size = min (vid.width, vid.height) + 2;

	// don't recalculate when only zooming in
	if (r_particles_init
	&&	r_particles.h_aspect	== h_aspect
	&&	r_particles.max_size	== max_size // >=
		)
		return;
	r_particles.h_aspect	= h_aspect;

	// free previously allocated memory before allocating
	if (r_particles_init)
	{
		for ( ; p <= r_particles.max_size ; p++)
		{
			free (r_particles.for_size[p]->scanwidth);
			free (r_particles.for_size[p]->offset);
			free (r_particles.for_size[p]);
		}
		p = 0; // the next 'for' loop will use this
		free (r_particles.for_size);
	}
	else
		r_particles_init = true;

	r_particles.max_size	= max_size; // must not be set before freeing the previously allocated memory

	r_particles.for_size = (d_particle_t **) malloc (sizeof (d_particle_t *) * (r_particles.max_size + 1));
	for ( ; p <= r_particles.max_size ; p++)
	{
		r_particles.for_size[p] = (d_particle_t *) malloc (sizeof (d_particle_t));

		pix = (double) (p + 2);
		width	= floor (0.5 + pix * r_particles.h_aspect);
		height	= floor (0.5 + pix);
		if (width < 1.0)
			width = 1.0;
		if (height < 1.0)
			height = 1.0;

		r_particles.for_size[p]->h_radius = (int) (0.5 * width );
		r_particles.for_size[p]->v_radius = (int) (0.5 * height);
		r_particles.for_size[p]->zcenter = r_particles.for_size[p]->h_radius + d_zwidth * r_particles.for_size[p]->v_radius;

		r_particles.for_size[p]->vrectright		= r_refdef.vrectright  - (int) width;
		r_particles.for_size[p]->vrectbottom	= r_refdef.vrectbottom - (int) height;

		count = height;
		r_particles.for_size[p]->lines = (int) count - 1;
		r_particles.for_size[p]->scanwidth	= (int *) malloc (sizeof (int) * (int) height);
		r_particles.for_size[p]->offset		= (int *) malloc (sizeof (int) * (int) height);
		for ( ; count > 0.0 ; count--)
		{
			// this "magic number" of 0.4 was achieved through trial and testing
			offset = 0.4 + (width - width * cos (asin (-1.0 + 2.0 * (count / height)))) * 0.5;
			r_particles.for_size[p]->scanwidth	[ (int) count - 1] = (int) (width - offset); // *2
			r_particles.for_size[p]->offset		[ (int) count - 1] = (int) offset;
		}
		r_particles.for_size[p]->unit = NULL;
	}
}
// Manoel Kasimier - end

void D_StartParticles (void)
{
	#if 0
	int p = 0;
	for ( ; p <= r_particles.max_size ; p++)
		r_particles.for_size[p]->unit = NULL;
	#endif
}

void D_DrawParticle_33 (particle_t *pparticle) // Manoel Kasimier
{
	vec3_t	local, transformed;
	float	zi;
	byte	*pdest;
	short	*pz;
	int		i, izi, u, v;
	double	pix, width, height; // Manoel Kasimier

	// transform point
	VectorSubtract (pparticle->org, r_origin, local);

	transformed[0] = DotProduct(local, r_pright);
	transformed[1] = DotProduct(local, r_pup);
	transformed[2] = DotProduct(local, r_ppn);

	if (transformed[2] < PARTICLE_Z_CLIP)
		return;

	// project the point
	// FIXME: preadjust xcenter and ycenter
	zi = 1.0 / transformed[2];

	izi = (int)(zi * 0x8000);

	// Manoel Kasimier - begin
	d_pix_shift = 8 - (int) ( ( ( (vid.aspect < 1.0) ? 1.0 : vid.aspect)* (float) r_refdef.vrect.width) / ( (1.0f / 320.0f) * (float) vid.width));
	pix = (double) ( (int) ( (float) izi * fovscale * ( (float) vid.width / 80.0f)) >> d_pix_shift);
	// doubled the minimum size here to make sure the particle will appear,
	// since the width and height calculations reduces its drawing area
	if (pix < 2.0)
		pix = 2.0;

	width	= floor (0.5 + pix * ((vid.aspect < 1.0) ? 1.0 : (1.0 / (double) vid.aspect)));
	height	= floor (0.5 + pix * ((vid.aspect < 1.0) ? (double) vid.aspect : 1.0));
	if (width < 1.0)
		width = 1.0;
	if (height < 1.0)
		height = 1.0;

	// width and height needs to be typecasted individually to avoid mismatched precision
	u = (int) (xcenter + transformed[0] * zi) - (int) (0.5 * width );
	v = (int) (ycenter - transformed[1] * zi) - (int) (0.5 * height);
	if	(
		(u > d_vrectright_particle	- (int) width)
	||	(v > d_vrectbottom_particle	- (int) height)
	||	(u < d_vrectx)
	||	(v < d_vrecty)
		)
		return;

	pdest	= u + d_viewbuffer	+ d_scantable[v];
	pz		= u + d_pzbuffer	+ (d_zwidth * v);

	if (pz[ (int) (width * 0.5) + d_zwidth * (int) (height * 0.5)] <= izi)
	{
		double
			scanwidth
		,	offset
		,	count = height
			;
		for ( ; count > 0.0 ; count--, pz += d_zwidth, pdest += screenwidth)
		{
			// this "magic number" of 0.4 was achieved through trial and testing
			offset = 0.4 + (width - width * cos (asin (-1.0 + 2.0 * (count / height)))) * 0.5;
			scanwidth = width - offset;
			for (i = (int) offset ; i < (int) scanwidth ; i++)
			{
			//	pz[i] = izi;
				pdest[i] = colorblendingmap[(int)(pparticle->color + pdest[i]*256)];
			}
		}
	}
	// Manoel Kasimier - end
}

void D_DrawParticle_66 (particle_t *pparticle) // Manoel Kasimier
{
	vec3_t	local, transformed;
	float	zi;
	byte	*pdest;
	short	*pz;
	int		i, izi, u, v;
	double	pix, width, height; // Manoel Kasimier

	// transform point
	VectorSubtract (pparticle->org, r_origin, local);

	transformed[0] = DotProduct(local, r_pright);
	transformed[1] = DotProduct(local, r_pup);
	transformed[2] = DotProduct(local, r_ppn);

	if (transformed[2] < PARTICLE_Z_CLIP)
		return;

	// project the point
	// FIXME: preadjust xcenter and ycenter
	zi = 1.0 / transformed[2];

	izi = (int)(zi * 0x8000);

	// Manoel Kasimier - begin
	d_pix_shift = 8 - (int) ( ( ( (vid.aspect < 1.0) ? 1.0 : vid.aspect)* (float) r_refdef.vrect.width) / ( (1.0f / 320.0f) * (float) vid.width));
	pix = (double) ( (int) ( (float) izi * fovscale * ( (float) vid.width / 80.0f)) >> d_pix_shift);
	// doubled the minimum size here to make sure the particle will appear,
	// since the width and height calculations reduces its drawing area
	if (pix < 2.0)
		pix = 2.0;

	width	= floor (0.5 + pix * ((vid.aspect < 1.0) ? 1.0 : (1.0 / (double) vid.aspect)));
	height	= floor (0.5 + pix * ((vid.aspect < 1.0) ? (double) vid.aspect : 1.0));
	if (width < 1.0)
		width = 1.0;
	if (height < 1.0)
		height = 1.0;

	// width and height needs to be typecasted individually to avoid mismatched precision
	u = (int) (xcenter + transformed[0] * zi) - (int) (0.5 * width );
	v = (int) (ycenter - transformed[1] * zi) - (int) (0.5 * height);
	if	(
		(u > d_vrectright_particle	- (int) width)
	||	(v > d_vrectbottom_particle	- (int) height)
	||	(u < d_vrectx)
	||	(v < d_vrecty)
		)
		return;

	pdest	= u + d_viewbuffer	+ d_scantable[v];
	pz		= u + d_pzbuffer	+ (d_zwidth * v);

	if (pz[ (int) (width * 0.5) + d_zwidth * (int) (height * 0.5)] <= izi)
	{
		double
			scanwidth
		,	offset
		,	count = height
			;
		for ( ; count > 0.0 ; count--, pz += d_zwidth, pdest += screenwidth)
		{
			// this "magic number" of 0.4 was achieved through trial and testing
			offset = 0.4 + (width - width * cos (asin (-1.0 + 2.0 * (count / height)))) * 0.5;
			scanwidth = width - offset;
			for (i = (int) offset ; i < (int) scanwidth ; i++)
			{
				pz[i] = izi;
			//	pdest[i] = alphamap[(int)(pdest[i] + pparticle->color*256)];
				pdest[i] = pparticle->color;
			}
		}
	}
	// Manoel Kasimier - end
}

// Manoel Kasimier - begin
// transforms and lists the particle
void D_DrawParticle_C (particle_t *pparticle) // D_DrawParticle - renamed to not conflict with the asm version
{
	vec3_t			local, transformed;
	float			zi, izi;
	int				pix;
	d_particle_t	*p;
//	float scale = -1.0f * (1.0f - ((cl.time - pparticle->start_time) / (pparticle->die - cl.time)));
//	scale = 2.0f - 2.0f * scale * scale;

	// transform point
	VectorSubtract (pparticle->org, r_origin, local);

	transformed[2] = DotProduct (local, r_ppn);
	if (transformed[2] < PARTICLE_Z_CLIP)
		return;

	transformed[0] = DotProduct (local, r_pright);
	transformed[1] = DotProduct (local, r_pup);

	// project the point
	zi = 1.0 / transformed[2];

	izi = zi * 0x8000;

	pix = ((int) (izi * r_particles.scale/* * scale*/) >> 8) - 2;
	if (pix < 0)
		p = r_particles.for_size[0];
	else if (pix < r_particles.max_size)
		p = r_particles.for_size[pix];
	else
		return;

	// width and height needs to be typecasted individually to avoid mismatched precision
	pparticle->u = (int) (xcenter + transformed[0] * zi) - p->h_radius;
	pparticle->v = (int) (ycenter - transformed[1] * zi) - p->v_radius;
	if	(
		(pparticle->u > p->vrectright)
	||	(pparticle->v > p->vrectbottom) // ok
	||	(pparticle->u < d_vrectx) // ok
	||	(pparticle->v < d_vrecty)
		)
		return;

	pparticle->z = (int) izi;
	pparticle->drawcolor = (byte) pparticle->color;

	pparticle->next_to_draw = p->unit;
	p->unit = pparticle;
}

//void D_DrawParticles (void)
void D_EndParticles (void)
{
	byte	*pdest;
	short	*pz;
	int		i, w, count, size = 0;
	d_particle_t	*p;
	particle_t		*pparticle;
	float alpha;

	for ( ; size <= r_particles.max_size ; size++)
	{
		p = r_particles.for_size[size];
		while (p->unit != NULL)
		{
			pparticle = p->unit;
			pdest	= d_viewbuffer	+ d_scantable[pparticle->v] + pparticle->u;
			pz		= d_pzbuffer	+ (d_zwidth * pparticle->v) + pparticle->u;

			if (pz[p->zcenter] <= pparticle->z)
			{
				count = p->lines;
				alpha = 1.0 - ((cl.time - pparticle->start_time) / (pparticle->die - cl.time)); // calcular isso no D_DrawParticle_C

				for ( ; count > -1 ; count-- , pz += d_zwidth , pdest += screenwidth)
				#if 0
				{
					w = p->scanwidth[count] - p->h_radius;
				//	memset (pz		+ p->offset[count], (int) pparticle->z			, (w << 1) * sizeof (*pz));
				//	memset (pdest	+ p->offset[count], (int) pparticle->drawcolor	, (w << 1));
					Q_memset (pz	+ p->offset[count], pparticle->z				, (w << 1) * sizeof (*pz));
					Q_memset (pdest	+ p->offset[count], (int) pparticle->drawcolor	, (w << 1));
				}
				#else
					for (i = p->offset[count] , w = p->scanwidth[count] ; i < w ; i++)
					{
						pz[i] = pparticle->z;
						pdest[i] = pparticle->drawcolor;
					}
				#endif
			}
			p->unit = pparticle->next_to_draw;
		}
		p->unit = NULL;
	}
}
// Manoel Kasimier - end
